import jax.numpy as jnp

from ..._misc import inexact_asarray
from ..._problem import AbstractBoundedMinimisation


class BOX2(AbstractBoundedMinimisation):
    """Box problem in 2 variables, obtained by fixing X3 = 1 in BOX3.

    This function is a nonlinear least squares with 10 groups.
    Each group has 2 nonlinear elements of exponential type.

    Source: Problem 11 in
    A.R. Buckley,
    "Test functions for unconstrained minimization",
    TR 1989CS-3, Mathematics, statistics and computing centre,
    Dalhousie University, Halifax (CDN), 1989.

    SIF input: Ph. Toint, Dec 1989.

    classification: SXR2-AN-3-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    n: int = 3  # 3 variables (including fixed x3)

    @property
    def y0(self):
        # Starting point from SIF file (x1, x2, x3)
        return jnp.array([0.0, 10.0, 1.0])

    @property
    def args(self):
        return ()

    def objective(self, y, args):
        """Compute the objective function.

        Minimize sum_{i=1}^{10} G(i)^2
        where G(i) = x3 * COEFF * (exp(-0.1*i*x1) - exp(-0.1*i*x2))
        and COEFF = -exp(-0.1*i) + exp(-i)

        Note: x3 is fixed at 1.0 but included in the variable vector
        """
        x1, x2, x3 = y

        # Vectorized implementation
        i_vals = jnp.arange(1, 11)  # i = 1 to 10
        ri = inexact_asarray(i_vals)
        mti = inexact_asarray(-0.1) * ri

        # Elements with parameter T = -0.1*i
        a_i = jnp.exp(mti * x1)
        b_i = jnp.exp(mti * x2)

        # COEFF = -exp(-0.1*i) + exp(-i) from SIF file
        coeff = -jnp.exp(mti) + jnp.exp(-ri)

        # Group values: G(i) = x3 * coeff + (a_i - b_i)
        g_i = x3 * coeff + a_i - b_i

        # Sum of squares
        return jnp.sum(g_i**2)

    @property
    def expected_result(self):
        # The optimal solution is not explicitly given in the SIF file
        return None

    @property
    def expected_objective_value(self):
        # From the SIF file comment
        return jnp.array(0.0)

    @property
    def bounds(self):
        # X1 and X2 are free, X3 is fixed at 1.0
        lower = jnp.array([-jnp.inf, -jnp.inf, 1.0])
        upper = jnp.array([jnp.inf, jnp.inf, 1.0])
        return lower, upper



import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class SPMSRTLS(AbstractUnconstrainedMinimisation):
    """Liu and Nocedal tridiagonal matrix square root least-squares problem.

    This is a least-squares variant of problem SPMSQRT. We seek a tridiagonal
    matrix X such that X^T * X approximates a given tridiagonal matrix A in
    the least-squares sense.

    Source: problem 151 (p. 93) in
    A.R. Buckley,
    "Test functions for unconstrained minimization",
    TR 1989CS-3, Mathematics, statistics and computing centre,
    Dalhousie University, Halifax (CDN), 1989.

    SIF input: Ph. Toint, Dec 1989.

    Classification: SUR2-AN-V-V
    """

    _m: int = 1667  # Default dimension
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self):
        """Number of variables: 3*M-2."""
        return 3 * self._m - 2

    @property
    def args(self):
        """No additional arguments."""
        return None

    def _compute_b_matrix(self):
        """Compute the entries of the tridiagonal matrix B."""
        m = self._m

        # Initialize B matrix entries
        b = {}

        # Row 1
        b[(1, 1)] = jnp.sin(1.0)
        b[(1, 2)] = jnp.sin(4.0)

        # Rows 2 to M-1
        k = 2
        for i in range(2, m):
            for j in [i - 1, i, i + 1]:
                k += 1
                rk = float(k)
                b[(i, j)] = jnp.sin(rk * rk)

        # Row M
        k += 1
        rk = float(k)
        b[(m, m - 1)] = jnp.sin(rk * rk)
        k += 1
        rk = float(k)
        b[(m, m)] = jnp.sin(rk * rk)

        return b

    def _compute_a_matrix(self, b):
        """Compute the pentadiagonal matrix A from B."""
        m = self._m
        a = {}

        # Diagonal entries
        a[(1, 1)] = b[(1, 1)] * b[(1, 1)] + b[(1, 2)] * b[(2, 1)]

        for i in range(2, m):
            a[(i, i)] = (
                b[(i, i)] * b[(i, i)]
                + b[(i - 1, i)] * b[(i, i - 1)]
                + b[(i + 1, i)] * b[(i, i + 1)]
            )

        a[(m, m)] = b[(m, m)] * b[(m, m)] + b[(m - 1, m)] * b[(m, m - 1)]

        # Sub-diagonal entries
        for i in range(1, m):
            a[(i + 1, i)] = (
                b[(i + 1, i)] * b[(i, i)] + b[(i + 1, i + 1)] * b[(i + 1, i)]
            )

        # Super-diagonal entries
        for i in range(2, m + 1):
            a[(i - 1, i)] = (
                b[(i - 1, i)] * b[(i, i)] + b[(i - 1, i - 1)] * b[(i - 1, i)]
            )

        # Sub-sub-diagonal entries
        for i in range(2, m):
            a[(i + 1, i - 1)] = b[(i + 1, i)] * b[(i, i - 1)]

        # Super-super-diagonal entries
        for i in range(2, m):
            a[(i - 1, i + 1)] = b[(i - 1, i)] * b[(i, i + 1)]

        return a

    @property
    def y0(self):
        """Initial guess: 0.2 * B matrix entries."""
        m = self._m
        b = self._compute_b_matrix()
        y0 = jnp.zeros(self.n)

        # X(1,1) and X(1,2)
        y0 = y0.at[0].set(0.2 * b[(1, 1)])
        y0 = y0.at[1].set(0.2 * b[(1, 2)])

        # X(i,i-1), X(i,i), X(i,i+1) for i = 2 to M-1
        idx = 2
        for i in range(2, m):
            y0 = y0.at[idx].set(0.2 * b[(i, i - 1)])
            y0 = y0.at[idx + 1].set(0.2 * b[(i, i)])
            y0 = y0.at[idx + 2].set(0.2 * b[(i, i + 1)])
            idx += 3

        # X(M,M-1) and X(M,M)
        y0 = y0.at[idx].set(0.2 * b[(m, m - 1)])
        y0 = y0.at[idx + 1].set(0.2 * b[(m, m)])

        return y0

    def objective(self, y, args):
        """Compute the least-squares objective function.

        The objective is the sum of squared differences between entries of
        A and X^T * X, where X is the tridiagonal matrix formed from y.
        """
        del args  # Not used

        m = self._m

        # Compute target matrix A
        b = self._compute_b_matrix()
        a = self._compute_a_matrix(b)

        # Extract X matrix entries from y
        x11 = y[0]
        x12 = y[1]

        # Initialize objective
        obj = 0.0

        # Line 1 of X^T * X
        # E(1,1): D(1) + G(1) where D(1) = x11^2, G(1) = x12 * x21
        # Since X is tridiagonal, x21 = x12
        e11 = x11**2 + x12 * x12
        obj += (e11 - a[(1, 1)]) ** 2

        # E(1,2): H(1) + R(1) where H(1) = x11 * x12, R(1) = x12 * x22
        idx = 2
        x22 = y[idx + 1] if m > 2 else y[idx]  # X(2,2)
        e12 = x11 * x12 + x12 * x22
        obj += (e12 - a.get((1, 2), 0.0)) ** 2

        # E(1,3): S(1) where S(1) = x12 * x23
        if m > 2:
            x23 = y[idx + 2]  # X(2,3)
            e13 = x12 * x23
            obj += (e13 - a.get((1, 3), 0.0)) ** 2

        # Lines 2 to M-1 of X^T * X
        idx = 2
        for i in range(2, m):
            # Extract X entries for row i
            xi_im1 = y[idx]  # X(i,i-1)
            xi_i = y[idx + 1]  # X(i,i)
            xi_ip1 = y[idx + 2] if i < m - 1 else 0.0  # X(i,i+1)

            # Also need entries from adjacent rows
            if i > 2:
                xim1_im2 = y[idx - 3]  # X(i-1,i-2)
                xim1_im1 = y[idx - 2]  # X(i-1,i-1)
                xim1_i = y[idx - 1]  # X(i-1,i)
            else:
                xim1_im2 = 0.0
                xim1_im1 = x11
                xim1_i = x12

            if i < m - 1:
                xip1_i = xi_ip1  # X(i+1,i) = X(i,i+1) by symmetry
                xip1_ip1 = y[idx + 4] if i < m - 2 else y[idx + 3]  # X(i+1,i+1)
                xip1_ip2 = y[idx + 5] if i < m - 2 else 0.0  # X(i+1,i+2)
            else:
                xip1_i = y[idx + 2]  # X(M,M-1)
                xip1_ip1 = y[idx + 3]  # X(M,M)
                xip1_ip2 = 0.0

            # E(i,i-2): A(i) = xi_im1 * xim1_im2
            if i > 2:
                e_iim2 = xi_im1 * xim1_im2
                obj += (e_iim2 - a.get((i, i - 2), 0.0)) ** 2

            # E(i,i-1): B(i) + C(i) where B(i) = xi_im1 * xim1_im1, C(i) = xi_i * xi_im1
            e_iim1 = xi_im1 * xim1_im1 + xi_i * xi_im1
            obj += (e_iim1 - a.get((i, i - 1), 0.0)) ** 2

            # E(i,i): F(i) + D(i) + G(i)
            # F(i) = xi_im1 * xim1_i, D(i) = xi_i^2, G(i) = xi_ip1 * xip1_i
            f_i = xi_im1 * xim1_i
            d_i = xi_i**2
            g_i = xi_ip1 * xip1_i if i < m else 0.0
            e_ii = f_i + d_i + g_i
            obj += (e_ii - a.get((i, i), 0.0)) ** 2

            # E(i,i+1): H(i) + R(i)
            # H(i) = xi_i * xi_ip1, R(i) = xi_ip1 * xip1_ip1
            if i < m:
                h_i = xi_i * xi_ip1
                r_i = xi_ip1 * xip1_ip1
                e_iip1 = h_i + r_i
                obj += (e_iip1 - a.get((i, i + 1), 0.0)) ** 2

            # E(i,i+2): S(i) = xi_ip1 * xip1_ip2
            if i < m - 1:
                s_i = xi_ip1 * xip1_ip2
                e_iip2 = s_i
                obj += (e_iip2 - a.get((i, i + 2), 0.0)) ** 2

            idx += 3

        # Line M of X^T * X
        xm_mm1 = y[-2]  # X(M,M-1)
        xm_m = y[-1]  # X(M,M)

        # Need X(M-1,M-2), X(M-1,M-1), X(M-1,M)
        if m > 2:
            xm1_mm2 = y[-5]  # X(M-1,M-2)
            xm1_mm1 = y[-4]  # X(M-1,M-1)
            xm1_m = y[-3]  # X(M-1,M)
        else:
            xm1_mm2 = 0.0
            xm1_mm1 = x11
            xm1_m = x12

        # E(M,M-2): A(M) = xm_mm1 * xm1_mm2
        if m > 2:
            em_mm2 = xm_mm1 * xm1_mm2
            obj += (em_mm2 - a.get((m, m - 2), 0.0)) ** 2

        # E(M,M-1): B(M) + C(M)
        em_mm1 = xm_mm1 * xm1_mm1 + xm_m * xm_mm1
        obj += (em_mm1 - a.get((m, m - 1), 0.0)) ** 2

        # E(M,M): F(M) + D(M)
        fm = xm_mm1 * xm1_m
        dm = xm_m**2
        em_m = fm + dm
        obj += (em_m - a.get((m, m), 0.0)) ** 2

        return obj

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value (not provided in SIF)."""
        return None



import jax.numpy as jnp
from jax import Array

from ..._misc import inexact_asarray
from ..._problem import AbstractNonlinearEquations


class CYCLOOCF(AbstractNonlinearEquations):
    """
    The cyclooctane molecule configuration problem (no fixed variables).

    The cyclooctane molecule is comprised of eight carbon atoms aligned
    in an equally spaced ring. When they take a position of minimum
    potential energy so that next-neighbours are equally spaced.

    Given positions v_1, ..., v_p in R^3 (with p = 8 for cyclooctane),
    and given a spacing c^2 we have that

       ||v_i - v_i+1,mod p||^2 = c^2 for i = 1,..,p, and
       ||v_i - v_i+2,mod p||^2 = 2p/(p-2) c^2

    This is a version of CYCLOOCT.SIF without the fixed variables.

    Source:
    an extension of the cyclooctane molecule configuration space as
    described in (for example)

     E. Coutsias, S. Martin, A. Thompson & J. Watson
     "Topology of cyclooctane energy landscape"
     J. Chem. Phys. 132-234115 (2010)

    SIF input: Nick Gould, Feb 2020.

    classification  NQR2-MN-V-V
    """

    p: int = 10000  # Number of molecules
    c: float = 1.0  # Radius
    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    @property
    def n(self) -> int:
        """Number of variables: Y2, Z2, and X3..XP, Y3..YP, Z3..ZP
        Total: 2 + 3*(P-2) = 3*P - 4
        """
        return 3 * self.p - 4

    def _get_positions(self, y: Array) -> Array:
        """Convert variable vector to full position array (p, 3)"""
        p = self.p
        positions = jnp.zeros((p, 3))

        # v_1 = (0, 0, 0) - fixed
        # v_2 = (0, Y2, Z2) where X2=0 is fixed
        positions = positions.at[1, 1].set(y[0])  # Y2
        positions = positions.at[1, 2].set(y[1])  # Z2

        # v_3 to v_p: X(i), Y(i), Z(i) for i = 3 to P
        # Variables y[2:] contain [X3, Y3, Z3, X4, Y4, Z4, ...]
        remaining_positions = y[2:].reshape(p - 2, 3)
        positions = positions.at[2:].set(remaining_positions)

        return positions

    def constraint(self, y: Array):
        """Compute the constraint equations for cyclooctane configuration"""
        p = self.p
        c2 = self.c**2
        sc2 = (2.0 * p / (p - 2)) * c2

        positions = self._get_positions(y)

        # Nearest neighbor constraints: ||v_i - v_{i+1}||^2 = c^2
        next_indices = (jnp.arange(p) + 1) % p
        diff_next = positions - positions[next_indices]
        dist_sq_next = jnp.sum(diff_next**2, axis=1)
        constraints_a = dist_sq_next - c2

        # Next-next neighbor constraints: ||v_i - v_{i+2}||^2 = 2p/(p-2)*c^2
        next2_indices = (jnp.arange(p) + 2) % p
        diff_next2 = positions - positions[next2_indices]
        dist_sq_next2 = jnp.sum(diff_next2**2, axis=1)
        constraints_b = dist_sq_next2 - sc2

        # Interleave constraints as A(1), B(1), A(2), B(2), ..., A(P), B(P)
        # This matches the SIF file constraint ordering
        eq_constraints = jnp.stack([constraints_a, constraints_b], axis=1).flatten()

        # No inequality constraints for this problem
        return eq_constraints, None

    @property
    def y0(self) -> Array:
        """Initial point: Y2, Z2 = 0, then i/P for molecule i"""
        p = self.p
        n = self.n

        # Initialize all variables
        initial = jnp.zeros(n)

        # Y2 and Z2 start at 0 (per SIF file START POINT section)
        # initial[0] and initial[1] are already 0

        # Vectorized initialization for X3..XP, Y3..YP, Z3..ZP
        # Create values for molecules 3 to p
        molecule_indices = jnp.arange(3, p + 1)
        values = molecule_indices / p

        # Each molecule has 3 coordinates (X, Y, Z) all set to the same value
        remaining_values = jnp.repeat(values, 3)

        # Set all remaining positions at once
        initial = initial.at[2:].set(remaining_values)

        return inexact_asarray(initial)

    @property
    def args(self):
        return None

    @property
    def expected_result(self) -> None:
        """Solution represents minimum energy configuration"""
        return None

    @property
    def bounds(self):
        """No bounds for this problem"""
        return None

    @property
    def expected_objective_value(self) -> Array:
        """Expected optimal objective value: 0"""
        return inexact_asarray(0.0)



import jax.numpy as jnp

from ..._problem import AbstractConstrainedMinimisation


# TODO: Human review needed - problem structure doesn't match pycutest
# The objective/constraint split needs to be reconsidered
# TODO: Human review needed - ANTWERP
# Attempts made:
# 1. Fixed constraint formulation (AGE2/AGE3 now correctly use N2/N3)
# 2. Initial value calculation follows SIF file formulas
# Suspected issues: Initial probability calculations may differ from pycutest defaults
# Additional resources needed: Understanding of pycutest initial value handling
class ANTWERP(AbstractConstrainedMinimisation):
    """Synthetic population estimation problem for Antwerp.

    This problem estimates the distribution of households in Belgian municipalities.
    Three household types are considered:
    - Type F: a couple + 1 to 5 children + 0 to 2 additional adults
    - Type W: a woman + 1 to 3 children + 0 to 2 additional adults
    - Type M: a man + 1 to 3 children + 0 to 2 additional adults

    The problem is noted to be very ill-conditioned.

    Source: L. Schoonbeek, "A household disaggregation procedure for the analysis
    of labour market policies", presented at the Franco-Belgian Symposium on
    "Desaggregation Methods in Socio-Economics", Brussels, 1986.

    SIF input: Nick Gould, February 1991

    Classification: SLR2-RN-27-8-0-3-24-0-2-0-8-0-0-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem data - demographic statistics for Antwerp
    # Number of households by size
    m3: float = 5827.0
    m4: float = 5929.0
    m5: float = 2889.0
    m6: float = 1065.0
    m7: float = 364.0
    m8: float = 180.0

    # Number of households by type and children
    m1f: float = 1974.0
    m2f: float = 5645.0
    m3f: float = 2813.0

    m1w: float = 2967.0
    m2w: float = 1095.0
    m3w: float = 296.0

    m1m: float = 3023.0
    m2m: float = 1133.0
    m3m: float = 327.0

    # Population counts
    n0: float = 27435.0  # Children (by type)
    n1: float = 3359.0  # Adults in W households
    n2: float = 4444.0  # Adults in M households
    n3: float = 14838.0  # Adults in F households
    n4: float = 18949.0  # Adults in F households (different age group)

    ninf: float = 48262.0  # Total individuals in F households
    ninn: float = 17784.0  # Total individuals in N households

    def __init__(self):
        """Initialize the ANTWERP problem."""
        # No fields to initialize as all data is defined at class level
        pass

    @property
    def n(self):
        """Total number of variables."""
        return 27

    @property
    def m(self):
        """Total number of constraints."""
        return 10  # 8 equality + 2 inequality

    def objective(self, y, args):
        """Compute the least-squares objective function."""
        del args

        # Extract variables
        p1f, p2f, p3f, p4f, p5f = y[0:5]
        p1w, p2w, p3w = y[5:8]
        p1m, p2m, p3m = y[8:11]
        q0f, q1f, q2f = y[11:14]
        q0w, q1w, q2w = y[14:17]
        q0m, q1m, q2m = y[17:20]
        nf, nw, nm = y[20:23]
        nc2, na2, nc3, na3 = y[23:27]

        # Compute household size predictions
        hsz3 = (p1f * q0f + p1f * q1f) * nf + p1w * q0w * nw + p1m * q0m * nm
        hsz4 = (
            (p2f * q0f + p1f * q2f + p2f * q1f) * nf
            + (p1w * q1w + p2w * q0w) * nw
            + (p1m * q1m + p2m * q0m) * nm
        )
        hsz5 = (
            (p3f * q0f + p2f * q2f + p3f * q1f) * nf
            + (p1w * q2w + p2w * q1w + p3w * q0w) * nw
            + (p1m * q2m + p2m * q1m + p3m * q0m) * nm
        )
        hsz6 = (
            (p4f * q0f + p3f * q2f + p4f * q1f) * nf
            + (p2w * q2w + p3w * q1w) * nw
            + (p2m * q2m + p3m * q1m) * nm
        )
        hsz7 = (
            (p5f * q0f + p4f * q2f + p5f * q1f) * nf + p3w * q2w * nw + p3m * q2m * nm
        )
        hsz8 = (p5f * q2f) * nf

        # Compute household type predictions
        hst1f = p1f * nf
        hst2f = p2f * nf
        hst3f = (p3f + p4f + p5f) * nf
        hst1w = p1w * nw
        hst2w = p2w * nw
        hst3w = p3w * nw
        hst1m = p1m * nm
        hst2m = p2m * nm
        hst3m = p3m * nm

        # Compute population predictions
        hch = (
            (p1f + 2 * p2f + 3 * p3f + 4 * p4f + 5 * p5f) * nf
            + (p1w + 2 * p2w + 3 * p3w) * nw
            + (p1m + 2 * p2m + 3 * p3m) * nm
            + nc2
            + nc3
        )
        had = (
            2 * nf
            + (q1f + 2 * q2f) * nf
            + (q1w + 2 * q2w) * nw
            + (q1m + 2 * q2m) * nm
            + na2
            + na3
        )
        hinf = nf * (3 + p1f + 2 * p2f + 3 * p3f + 4 * p4f + 5 * p5f + q1f + 2 * q2f)
        hinn = (
            nw * (1 + p1w + 2 * p2w + 3 * p3w + q1w + 2 * q2w)
            + nm * (1 + p1m + 2 * p2m + 3 * p3m + q1m + 2 * q2m)
            + nc2
            + na2
            + nc3
            + na3
        )

        # Compute squared errors (scaled by inverse of RHS)
        obj = 0.0
        obj += ((hsz3 - self.m3) / self.m3) ** 2
        obj += ((hsz4 - self.m4) / self.m4) ** 2
        obj += ((hsz5 - self.m5) / self.m5) ** 2
        obj += ((hsz6 - self.m6) / self.m6) ** 2
        obj += ((hsz7 - self.m7) / self.m7) ** 2
        obj += ((hsz8 - self.m8) / self.m8) ** 2

        obj += ((hst1f - self.m1f) / self.m1f) ** 2
        obj += ((hst2f - self.m2f) / self.m2f) ** 2
        obj += ((hst3f - self.m3f) / self.m3f) ** 2
        obj += ((hst1w - self.m1w) / self.m1w) ** 2
        obj += ((hst2w - self.m2w) / self.m2w) ** 2
        obj += ((hst3w - self.m3w) / self.m3w) ** 2
        obj += ((hst1m - self.m1m) / self.m1m) ** 2
        obj += ((hst2m - self.m2m) / self.m2m) ** 2
        obj += ((hst3m - self.m3m) / self.m3m) ** 2

        obj += ((hch - self.n0) / self.n0) ** 2
        nad_total = self.n1 + self.n2 + self.n3 + self.n4
        obj += ((had - nad_total) / nad_total) ** 2
        obj += ((hinf - self.ninf) / self.ninf) ** 2
        obj += ((hinn - self.ninn) / self.ninn) ** 2

        return obj

    def constraint(self, y):
        """Compute equality and inequality constraints."""
        # Extract variables
        p1f, p2f, p3f, p4f, p5f = y[0:5]
        p1w, p2w, p3w = y[5:8]
        p1m, p2m, p3m = y[8:11]
        q0f, q1f, q2f = y[11:14]
        q0w, q1w, q2w = y[14:17]
        q0m, q1m, q2m = y[17:20]
        nf, nw, nm = y[20:23]
        nc2, na2, nc3, na3 = y[23:27]

        # Equality constraints
        eq_constraints = []

        # AGE2: NC2 + NA2 = N2
        eq_constraints.append(nc2 + na2 - self.n2)

        # AGE3: NC3 + NA3 = N3
        eq_constraints.append(nc3 + na3 - self.n3)

        # PSF: P1F + P2F + P3F + P4F + P5F = 1
        eq_constraints.append(p1f + p2f + p3f + p4f + p5f - 1.0)

        # PSW: P1W + P2W + P3W = 1
        eq_constraints.append(p1w + p2w + p3w - 1.0)

        # PSM: P1M + P2M + P3M = 1
        eq_constraints.append(p1m + p2m + p3m - 1.0)

        # QSF: Q0F + Q1F + Q2F = 1
        eq_constraints.append(q0f + q1f + q2f - 1.0)

        # QSW: Q0W + Q1W + Q2W = 1
        eq_constraints.append(q0w + q1w + q2w - 1.0)

        # QSM: Q0M + Q1M + Q2M = 1
        eq_constraints.append(q0m + q1m + q2m - 1.0)

        # Inequality constraints (>= 0)
        ineq_constraints = []

        # INEQ2: NA2 - 2642 >= 0
        ineq_constraints.append(na2 - 2642.0)

        # INEQ3: NA3 >= 0
        ineq_constraints.append(na3)

        return jnp.array(eq_constraints), jnp.array(ineq_constraints)

    @property
    def y0(self):
        """Initial guess for variables."""
        # Compute initial values based on data
        sp1f = self.m1f / (self.m1f + self.m2f + self.m3f)
        sp2f = self.m2f / (self.m1f + self.m2f + self.m3f)
        sp1w = self.m1w / (self.m1w + self.m2w + self.m3w)
        sp2w = self.m2w / (self.m1w + self.m2w + self.m3w)
        sp3w = self.m3w / (self.m1w + self.m2w + self.m3w)
        sp1m = self.m1m / (self.m1m + self.m2m + self.m3m)
        sp2m = self.m2m / (self.m1m + self.m2m + self.m3m)
        sp3m = self.m3m / (self.m1m + self.m2m + self.m3m)

        snf = self.m1f + self.m2f + self.m3f
        snw = self.m1w + self.m2w + self.m3w
        snm = self.m1m + self.m2m + self.m3m

        return jnp.array(
            [
                sp1f,  # P1F
                sp2f,  # P2F
                0.15,  # P3F
                0.10,  # P4F
                0.05,  # P5F
                sp1w,  # P1W
                sp2w,  # P2W
                sp3w,  # P3W
                sp1m,  # P1M
                sp2m,  # P2M
                sp3m,  # P3M
                0.6,  # Q0F
                0.3,  # Q1F
                0.1,  # Q2F
                0.6,  # Q0W
                0.3,  # Q1W
                0.1,  # Q2W
                0.6,  # Q0M
                0.3,  # Q1M
                0.1,  # Q2M
                snf,  # NF
                snw,  # NW
                snm,  # NM
                0.0,  # NC2
                self.n2,  # NA2
                0.0,  # NC3
                self.n3,  # NA3
            ]
        )

    @property
    def bounds(self):
        """Get variable bounds."""
        # All probabilities in [0, 1], except P3F, P4F, P5F with tighter bounds
        # NF, NW, NM in [0, 10000]
        # NC2, NC3 are fixed at 0
        # NA2, NA3 are free
        lower = jnp.array(
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,  # P1F-P5F
                0.0,
                0.0,
                0.0,  # P1W-P3W
                0.0,
                0.0,
                0.0,  # P1M-P3M
                0.0,
                0.0,
                0.0,  # Q0F-Q2F
                0.0,
                0.0,
                0.0,  # Q0W-Q2W
                0.0,
                0.0,
                0.0,  # Q0M-Q2M
                0.0,
                0.0,
                0.0,  # NF, NW, NM
                0.0,
                -jnp.inf,  # NC2, NA2
                0.0,
                -jnp.inf,  # NC3, NA3
            ]
        )

        upper = jnp.array(
            [
                1.0,
                1.0,
                0.25,
                0.2,
                0.1,  # P1F-P5F
                1.0,
                1.0,
                1.0,  # P1W-P3W
                1.0,
                1.0,
                1.0,  # P1M-P3M
                1.0,
                1.0,
                1.0,  # Q0F-Q2F
                1.0,
                1.0,
                1.0,  # Q0W-Q2W
                1.0,
                1.0,
                1.0,  # Q0M-Q2M
                10000.0,
                10000.0,
                10000.0,  # NF, NW, NM
                0.0,
                jnp.inf,  # NC2, NA2
                0.0,
                jnp.inf,  # NC3, NA3
            ]
        )

        return (lower, upper)

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution (not provided in SIF)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value (not provided in SIF)."""
        return None

    def equality_constraints(self):
        """Mark which constraints are equalities."""
        # First 8 are equalities, last 2 are inequalities
        return jnp.array([True, True, True, True, True, True, True, True, False, False])



import jax.numpy as jnp
from jaxtyping import Array, Float

from ..._problem import AbstractBoundedMinimisation


class BDEXP(AbstractBoundedMinimisation):
    """BDEXP problem.

    A banded exponential problem.

    Source: Problem 56 in
    A.R. Conn, N.I.M. Gould, M. Lescrenier and Ph.L. Toint,
    "Performance of a multifrontal scheme for partially separable
    optimization",
    Report 88/4, Dept of Mathematics, FUNDP (Namur, B), 1988.

    SIF input: Ph. Toint, Dec 1989.

    classification OBR2-AY-V-0
    """

    # Default parameter
    N: int = 5000  # Number of variables

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    def objective(self, y, args):
        """Compute the objective function."""
        del args

        x = y

        # Vectorized computation of DEXP elements
        # Element A(i) uses variables X(i), X(i+1), X(i+2)
        v1 = x[:-2]  # x[0] to x[n-3]
        v2 = x[1:-1]  # x[1] to x[n-2]
        v3 = x[2:]  # x[2] to x[n-1]

        # From DEXP element:
        # U1 = V1 + V2
        # U2 = -V3 (note the negative sign from SIF: R  U2  V3  -1.0)
        u1 = v1 + v2
        u2 = -v3

        # F = exp(U1 * U2) * U1
        u1_times_u2 = u1 * u2
        exp_u1u2 = jnp.exp(u1_times_u2)

        # Sum all elements
        return jnp.sum(exp_u1u2 * u1)

    @property
    def y0(self):
        """Initial guess for variables."""
        return jnp.ones(self.N)

    @property
    def args(self):
        """Additional arguments (none for this problem)."""
        return None

    @property
    def expected_result(self):
        """Expected optimal solution."""
        return None

    @property
    def expected_objective_value(self):
        """Expected optimal objective value."""
        # From SIF file: SOLTN = 0.0
        return jnp.array(0.0)

    @property
    def n(self):
        """Number of variables."""
        return self.N

    @property
    def bounds(self) -> tuple[Float[Array, "n"], Float[Array, "n"]]:
        """Lower and upper bounds for variables.

        All variables have lower bound 0.0 and no upper bound.
        """
        lower = jnp.zeros(self.N)
        upper = jnp.full(self.N, jnp.inf)
        return lower, upper



import equinox as eqx

# TODO: Human review needed
# Attempts made: [fixed dtype promotion error, but significant gradient
#                 discrepancies remain]
# Suspected issues: [boundary condition implementation, discretization scheme,
#                    max gradient difference 0.41, last elements inconsistent]
# Additional resources needed: [verification of PDE discretization, boundary handling]
import jax.numpy as jnp

from ..._problem import AbstractUnconstrainedMinimisation


class BRATU1D(AbstractUnconstrainedMinimisation):
    """Bratu's problem in one dimension, according to Osborne.

    Source: Problem 121 (p. 99) in
    A.R. Buckley,
    "Test functions for unconstrained minimization",
    TR 1989CS-3, Mathematics, statistics and computing centre,
    Dalhousie University, Halifax (CDN), 1989.

    SIF input: Ph. Toint, Dec 1989.

    classification OXR2-MN-V-0
    """

    y0_iD: int = 0
    provided_y0s: frozenset = frozenset({0})

    # Problem parameter
    lambda_param: float = -3.4

    # Number of variables (must be odd)
    n: int = eqx.field(default=5001, init=False)

    def objective(self, y, args):
        """Compute the objective function (vectorized)."""
        n = self.n
        h = 1.0 / (n + 1)
        lambda_h = self.lambda_param * h
        two_lambda_h = 2.0 * lambda_h
        two_over_h = 2.0 / h

        # Variables include fixed boundary values x[0] = 0, x[n+1] = 0
        # but we only work with interior points x[1] to x[n]
        x = jnp.concatenate([jnp.array([0.0]), y, jnp.array([0.0])])

        # Vectorized computation of all terms
        # GA terms: 2/h * x[i]^2 for i=1 to n
        ga_terms = two_over_h * y * y

        # GB terms: -2/h * x[i] * x[i-1] for i=1 to n
        gb_terms = -two_over_h * x[1:-1] * x[:-2]

        # GC terms: BRA(x[i], x[i+1]) for i=0 to n
        # Compute differences
        d = x[1:] - x[:-1]

        # Compute exponentials
        exp_x = jnp.exp(x)

        # BRA element: (exp(y) - exp(x)) / (y - x)
        # Use where to handle the case d == 0 (though it shouldn't happen)
        bra_vals = jnp.where(
            d != 0.0,
            (exp_x[1:] - exp_x[:-1]) / d,
            exp_x[:-1],  # limit as d->0 is exp(x)
        )

        gc_terms = two_lambda_h * bra_vals

        # Sum all contributions
        obj = jnp.sum(ga_terms) + jnp.sum(gb_terms) + jnp.sum(gc_terms)

        return obj

    @property
    def y0(self):
        """Starting point."""
        n = self.n
        h = 1.0 / (n + 1)

        # Vectorized computation
        i_values = jnp.arange(1, n + 1, dtype=jnp.float64)
        x0 = -0.1 * h * (i_values * i_values)

        return x0

    @property
    def args(self):
        return None

    @property
    def expected_result(self):
        """Expected solution (not provided)."""
        return None

    @property
    def expected_objective_value(self):
        """Expected objective value for n=5001."""
        # From the SIF file comments:
        # n=11: -8.49454553
        # n=75: -8.51831187
        # n=101: -8.51859
        # n=501: -8.51892
        # n=1001: -8.51893
        # For n=5001, we can extrapolate
        return jnp.array(-8.51893)
